package ctbrec.sites.manyvids;

import static ctbrec.io.HttpConstants.*;
import static ctbrec.sites.manyvids.MVLive.*;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.UUID;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Objects;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.squareup.moshi.Moshi;

import ctbrec.Config;
import ctbrec.io.HttpException;
import ctbrec.sites.manyvids.wsmsg.JoinChat;
import ctbrec.sites.manyvids.wsmsg.Message;
import ctbrec.sites.manyvids.wsmsg.Ping;
import ctbrec.sites.manyvids.wsmsg.SendMessage;
import okhttp3.Cookie;
import okhttp3.Request;
import okhttp3.Response;
import okhttp3.WebSocket;
import okhttp3.WebSocketListener;
import okio.ByteString;

public class MVLiveClient {

    private static final Logger LOG = LoggerFactory.getLogger(MVLiveClient.class);

    private static MVLiveClient instance;
    private MVLive site;
    private WebSocket ws;
    private Moshi moshi;
    private Random rng = new Random();
    private volatile boolean running = false;

    private Cache<Integer, MVLiveModel> models = CacheBuilder.newBuilder().maximumSize(4000).build();
    private Lock lock = new ReentrantLock();
    private volatile boolean connecting = false;

    private MVLiveClient() {
        moshi = new Moshi.Builder().build();
    }

    public static synchronized MVLiveClient getInstance() {
        if (instance == null) {
            instance = new MVLiveClient();
        }
        return instance;
    }

    public void setSite(MVLive site) {
        this.site = site;
    }

    public void start() throws IOException {
        running = true;


        // make a call to the website to get necessary cookies
        Request req = new Request.Builder()
                .url(site.getBaseUrl())
                .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                .build();
        try (Response response = site.getHttpClient().execute(req)) {
            if (!response.isSuccessful()) {
                throw new HttpException(response.code(), response.message());
            }
        }

        int randomNumber = 100 + rng.nextInt(800);
        String randomString = UUID.randomUUID().toString().replace("-", "").substring(0, 8);
        String roomNumber = "176";
        String wsUrl = String.format("%s/api/%s/eventbus/%s/%s/websocket", WS_URL, roomNumber, randomNumber, randomString);
        List<Cookie> cookies = site.getHttpClient().getCookiesByName("PHPSESSID"); //, "XSRF-TOKEN");
        String cookieHeaderValue = cookies.stream().map(Object::toString).collect(Collectors.joining("; "));

        Thread watchDog = new Thread(() -> {
            while (running) {
                if (ws == null && !connecting) {
                    LOG.info("Websocket is null. Starting a new connection to {}", wsUrl);
                    Request websocketRequest = new Request.Builder()
                            .url(wsUrl)
                            .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                            .header(ORIGIN, WS_URL)
                            .header(COOKIE, cookieHeaderValue)
                            .build();
                    ws = createWebSocket(websocketRequest);
                }

                try {
                    Thread.sleep(10000);
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                    LOG.error("WatchDog couldn't sleep", e);
                    stop();
                    running = false;
                }
            }
        });
        watchDog.setDaemon(true);
        watchDog.setName("MVLive WebSocket WatchDog");
        watchDog.setPriority(Thread.MIN_PRIORITY);
        watchDog.start();
    }

    public void stop() {
        running = false;
        ws.close(1000, "Good Bye"); // terminate normally (1000)
    }

    public List<MVLiveModel> getModels() {
        lock.lock();
        try {
            LOG.trace("Models: {}", models.size());
            return new ArrayList<>(this.models.asMap().values());
        } finally {
            lock.unlock();
        }
    }

    private WebSocket createWebSocket(Request req) {
        connecting = true;
        WebSocket websocket = site.getHttpClient().newWebSocket(req, new WebSocketListener() {
            @Override
            public void onOpen(WebSocket webSocket, Response response) {
                super.onOpen(webSocket, response);
                try {
                    connecting = false;
                    models.invalidateAll();
                    LOG.debug("WS open: [{}]", response.body().string());
                } catch (IOException e) {
                    LOG.error("Error while processing onOpen event", e);
                }
            }

            @Override
            public void onClosed(WebSocket webSocket, int code, String reason) {
                super.onClosed(webSocket, code, reason);
                connecting = false;
                LOG.info("MVLive websocket closed: {} {}", code, reason);
                MVLiveClient.this.ws = null;
                if (!running) {
                    site.getHttpClient().shutdown();
                }
            }

            @Override
            public void onFailure(WebSocket webSocket, Throwable t, Response response) {
                super.onFailure(webSocket, t, response);
                connecting = false;
                if (response != null) {
                    int code = response.code();
                    String message = response.message();
                    LOG.error("MVLive websocket failure: {} {}", code, message, t);
                    response.close();
                } else {
                    LOG.error("MVLive websocket failure", t);
                }
                MVLiveClient.this.ws = null;
            }

            @Override
            public void onMessage(WebSocket webSocket, String text) {
                super.onMessage(webSocket, text);
                //msgBuffer.append(text);
                LOG.debug("Message: {}", text);
                text = Optional.ofNullable(text).orElse("");
                if (Objects.equal("o", text)) {
                    sendMessages(new Ping());
                    String roomId = "3d9b273a-8039-41f7-a5d8-bdd6972d697c";
                    String showId = "HARIBOO";
                    sendMessages(new JoinChat(roomId, showId, (m, r) -> {
                        System.out.println(m);
                        System.out.println(r);
                    }));
                } else if (text.startsWith("a")) {
                    JSONArray jsonArray = new JSONArray(text.substring(1));
                    for (int i = 0; i < jsonArray.length(); i++) {
                        String respJson = jsonArray.getString(i);
                        JSONObject response = new JSONObject(respJson);
                        String address = response.optString("address");
                        if (!address.isBlank()) {
                            SendMessage sendMessage = futureResponses.get(address);
                            if (sendMessage != null) {
                                sendMessage.handleResponse(response);
                            }
                        }
                    }
                }
            }

            Map<String, SendMessage> futureResponses = new HashMap<>();

            private void sendMessages(Message... messages) {
                JSONArray msgs = new JSONArray();
                for (Message msg : messages) {
                    msgs.put(msg.toString());
                    if(msg instanceof SendMessage) {
                        SendMessage sendMessage = (SendMessage) msg;
                        futureResponses.put(sendMessage.getReplyAddress(), sendMessage);
                    }
                }
                ws.send(msgs.toString());
            }

            @Override
            public void onMessage(WebSocket webSocket, ByteString bytes) {
                super.onMessage(webSocket, bytes);
                LOG.debug("Binary Message: {}", bytes.hex());
            }
        });
        return websocket;
    }


}
